/*
	SPDX-License-Identifier: MIT
	Copyright © 2023-2025 Amebis
*/

#pragma once

#include "common.hpp"
#include <CoreFoundation/CoreFoundation.h>
#include <memory>
#include <system_error>

namespace macstd {
	///
	/// Core Foundation object traits
	///
	template <typename T>
	struct cfobject_traits
	{
		static inline constexpr T invalid = static_cast<T>(NULL);

		///
		/// Releases a Core Foundation object
		///
		/// \sa [CFRelease function](https://developer.apple.com/documentation/corefoundation/1521153-cfrelease)
		///
		static void free(T h)
		{
			CFRelease(h);
		}

	private:
		static T duplicate(T h);
	};

	///
	/// Core Foundation object traits
	///
	template <typename T>
	using cfobject = handle<T, cfobject_traits<T>>;

	using cfarray = cfobject<CFArrayRef>;
	using cfdata = cfobject<CFDataRef>;
	using cfdictionary = cfobject<CFDictionaryRef>;
	using cferror = cfobject<CFErrorRef>;
	using cfmutabledata = cfobject<CFMutableDataRef>;
	using cfreadstream = cfobject<CFReadStreamRef>;
	using cfurl = cfobject<CFURLRef>;

	template <typename TR = std::char_traits<char>, typename A = std::allocator<char>>
	void strcat(std::basic_string<char, TR, A>& str, CFStringRef h, CFStringEncoding encoding = kCFStringEncodingUTF8)
	{
		auto ptr = CFStringGetCStringPtr(h, encoding);
		if (ptr) {
			str += ptr;
			return;
		}
		char buf[0x200];
		if (CFStringGetCString(h, buf, sizeof(buf)/sizeof(*buf), encoding)) {
			str += buf;
			return;
		}
		auto n = str.size();
		auto m = CFStringGetMaximumSizeForEncoding(CFStringGetLength(h), encoding);
		str.resize(n + m);
		if (!CFStringGetCString(h, &str[n], m, encoding)) [[unlikely]]
			throw std::runtime_error("CFStringGetCString failed");
		str.resize(n + strlen(&str[n]));
	}

	template <typename TR = std::char_traits<char>, typename A = std::allocator<char>>
	void strcpy(std::basic_string<char, TR, A>& str, CFStringRef h, CFStringEncoding encoding = kCFStringEncodingUTF8)
	{
		str.clear();
		strcat(str, h, encoding);
	}

	template <typename TR = std::char_traits<char>, typename A = std::allocator<char>>
	std::basic_string<char, TR, A> to_string(CFStringRef h, CFStringEncoding encoding = kCFStringEncodingUTF8)
	{
		std::basic_string<char, TR, A> str;
		strcat(str, h, encoding);
		return str;
	}

	///
	/// Core Foundation string
	///
	class cfstring : public cfobject<CFStringRef>
	{
	public:
		cfstring(CFStringRef h) : cfobject<CFStringRef>(h) {}

		cfstring(const char *s, CFStringEncoding encoding = kCFStringEncodingUTF8) : cfobject<CFStringRef>(CFStringCreateWithCString(kCFAllocatorDefault, s, encoding))
		{
			if (!valid()) [[unlikely]]
				throw std::bad_alloc();
		}

		template <typename TR, typename A>
		cfstring(const std::basic_string<char, TR, A> &s, CFStringEncoding encoding = kCFStringEncodingUTF8) : cfstring(s.c_str(), encoding) {}
	};

	///
	/// Core Foundation number
	///
	class cfnumber : public cfobject<CFNumberRef>
	{
	public:
		cfnumber(CFNumberRef h) : cfobject<CFNumberRef>(h) {}

		cfnumber(float n) : cfobject<CFNumberRef>(CFNumberCreate(kCFAllocatorDefault, kCFNumberFloatType, &n))
		{
			if (!valid()) [[unlikely]]
				throw std::bad_alloc();
		}

		cfnumber(double n) : cfobject<CFNumberRef>(CFNumberCreate(kCFAllocatorDefault, kCFNumberDoubleType, &n))
		{
			if (!valid()) [[unlikely]]
				throw std::bad_alloc();
		}
	};

	///
	/// Core Foundation runtime error
	///
	class cf_runtime_error : public num_runtime_error<CFIndex>
	{
	protected:
		cferror m_error;

	public:
		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		///
		cf_runtime_error(CFErrorRef error) :
			num_runtime_error<CFIndex>(CFErrorGetCode(error), message(error)),
			m_error(error)
		{
		}

		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		/// \param[in] msg    Error message
		///
		cf_runtime_error(CFErrorRef error, const std::string& msg) :
			num_runtime_error<CFIndex>(CFErrorGetCode(error), msg + ": " + message(error)),
			m_error(error)
		{
		}

		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		/// \param[in] msg    Error message
		///
		cf_runtime_error(CFErrorRef error, const char* msg) : cf_runtime_error(error, std::string(msg)) {}

		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		///
		cf_runtime_error(cferror&& error) :
			num_runtime_error<CFIndex>(CFErrorGetCode(error), message(error)),
			m_error(std::move(error))
		{
		}

		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		/// \param[in] msg    Error message
		///
		cf_runtime_error(cferror&& error, const std::string& msg) :
			num_runtime_error<CFIndex>(CFErrorGetCode(error), msg + ": " + message(error)),
			m_error(std::move(error))
		{
		}

		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		/// \param[in] msg    Error message
		///
		cf_runtime_error(cferror&& error, const char* msg) : cf_runtime_error(std::move(error), std::string(msg)) {}

		CFErrorRef cferror() const noexcept { return m_error; }

	protected:
		///
		/// Returns a string explaining the meaning of a Core Foundation error
		///
		static std::string message(CFErrorRef error)
		{
			std::string msg;
			cfstring cfstr(CFErrorCopyDescription(error));
			if (cfstr.valid() && CFStringGetLength(cfstr) > 0) {
				strcpy(msg, cfstr, kCFStringEncodingUTF8);
			} else {
				msg = "Core Foundation error ";
				msg += std::to_string(CFErrorGetCode(error));
			}
			cfstring domain(CFErrorGetDomain(error));
			if (domain.valid() && CFStringGetLength(domain) > 0) {
				msg += ", domain ";
				strcat(msg, domain, kCFStringEncodingUTF8);
			}
			return msg;
		}
	};
}
